import unittest
from scr.main import getArgs
from contextlib import redirect_stderr
from io import StringIO

class TestArgs(unittest.TestCase):
	def setUp(self):
		self.parser = getArgs()
		self.buf = StringIO()
		redirect_stderr(self.buf).__enter__()
	def test_standard(self):
		args = self.parser.parse_args(["input.pdf", "--start", "22", "index.txt", "output.pdf"])
		self.assertEqual(args.input_pdf_file, "input.pdf")
		self.assertEqual(args.first_page, 22)
		self.assertEqual(args.outline_file, "index.txt")
		self.assertEqual(args.output_pdf_file, "output.pdf")
	def test_dry(self):
		# test the dry flag
		args = self.parser.parse_args(["input.pdf", "--start", "12", "index.txt", "output.pdf", "--dry"])
		self.assertTrue(args.dry)
	def test_debug(self):
		# test the debug flag
		args = self.parser.parse_args(["input.pdf", "--start", "12", "index.txt", "output.pdf", "--debug"])
		self.assertTrue(args.debug)
	def test_missing_args(self):
		# test what happens when some required arguments are missing
		with self.assertRaises(SystemExit) as exc:
			self.parser.parse_args(["input.pdf", "--start", "12", "index.txt"])
			self.assertEqual(exc.exception.code, 2)
			self.assertIn("required", exc.exception.args[0])
	def test_wrong_args(self):
		# test what happens when the arguments are not in the expected format
		with self.assertRaises(SystemExit) as exc:
			self.parser.parse_args(["input.pdf", "--start", "ciao", "index.txt", "output.pdf"])
			self.assertEqual(exc.exception.code, 2)
			self.assertIn("invalid int value", exc.exception.args[0])
	def test_negative_start(self):
		# test what happens when the start argument is negative
		with self.assertRaises(Exception) as exc:
			self.parser.parse_args(["input.pdf", "--start", "-1", "index.txt", "output.pdf"])
			self.assertEqual(exc.exception.code, 2)
			self.assertIn("the first real page number must be greater than or equal to 1", exc.exception.args[0])
	# TODO: handle extract arguments

if __name__ == '__main__':
	unittest.main()


import unittest
from scr.main import addLogicNums
from types import SimpleNamespace
import pikepdf

class TestLogicNumbers(unittest.TestCase):
	def setUp(self):
		self.args = SimpleNamespace(debug=False, dry=False)
		self.start = 1
	def writeNums(self, file, ref=None):
		# write logical page numbers to the pdf file and compare them with the reference if provided
		with pikepdf.open(file) as pdf:
			addLogicNums(pdf, self.start)
			if ref is not None:
				with pikepdf.open(ref) as ref:
					self.assertEqual(pdf.Root.PageLabels, ref.Root.PageLabels)
			pdf.save('/tmp/out.pdf')
	def test_onlyRealPages(self):
		# start=1
		self.writeNums("tests/sample1/input.pdf", "tests/sample1/pdf_ref.pdf")
	def test_nonExistingPage(self):
		# start>last page number
		with self.assertRaises(IndexError) as exc:
			self.start = 17
			self.writeNums("tests/sample7/input.pdf")
	def test_standard(self):
		self.start = 2
		self.writeNums("tests/sample5/input.pdf", "tests/sample5/pdf_shifted_ref.pdf")
	def test_fromExisting(self):
		# write on a pdf that already has logical page numbers, with a different start
		self.writeNums("tests/sample4/input.pdf", "tests/sample4/pdf_ref.pdf")
	
if __name__ == '__main__':
	unittest.main()


#!/usr/bin/env python3
import pikepdf
from pikepdf import Name, Dictionary, Array
from outline_model import OutlineElement, getNumber
import argparse
import re

def getArgs():
    parser = argparse.ArgumentParser(
        prog="PDF_outline_add",
        description="Add PDF outline and logical page labels"
    )

    parser.add_argument("input_pdf_file", help="Path to input PDF")
    parser.add_argument("outline_file", help="Path to outline text file")
    parser.add_argument("output_pdf_file", help="Path to output PDF")
    parser.add_argument("-s", "--start", action="store", dest="first_page", type=int, help="First real page number (1-based)", required=False, default=1)

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug output"
    )
    parser.add_argument(
        "-d",
        "--dry",
        action="store_true",
        help="Print the parsed index"
    )

    return parser

def parseOutline(file_outline, start=1, args=None):
    r_entry = r"^(\s*)(([ivxlcdmIVXLCDM]||\d)+)\s+(.*?)\s*$"          # Outline entry regex 
    outline_items = []
    with open(file_outline, 'r') as f:
        lines = [line.rstrip() for line in f.readlines() if not (line == '' or line == '\n' or line == '\r')]
        prev = 0
        par = None
        syntax = re.compile(r_entry)
        for line in lines:
            parts = syntax.match(line).groups()
            if len(parts) >= 3:
                title = parts[3]
                page_number = getNumber(parts[1], start)
                level = int(parts[0].count(' '))
                if args and args.debug:
                    print("title: %s, page number: %s, level: %s, prev: %s" % (title, page_number, level, prev))
                if level == 0:
                    outline_items.append(OutlineElement(title, level, page_number))
                    par = outline_items[-1]
                elif prev - level < 2 or level <= prev:
                    while(prev >= level):
                        prev -= 1
                        par = par.parent
                    par.add_child(OutlineElement(title, level, page_number, par))
                    par = par.children[-1]
                else:
                    raise Exception("Error: the difference between the next subsection level and this one must not be bigger than one: page title: %s, page number: %s, level: %s\n%s" % (title, page_number, level, prev))
                prev = level
            else:
                raise Exception("Error: line does not match the expected format: %s" % line)

    # stampo l'indice parsato (debug)
    if args and args.dry:
        for item in outline_items:
            print(repr(item))

    return outline_items

def addLogicNums(pdf, start):
    # inserisco numerazione logica con numeri romani
    # Creiamo la struttura dei numeri di pagina (PageLabels)
    # /Nums è un array dove ogni coppia è: [indice_pagina_inizio, dizionario_stile]
    # L'indice parte da 0 (0 = prima pagina del PDF)
    if start > len(pdf.pages):
        raise IndexError("Error: the first real page number is bigger than the total number of pages in the PDF: %s > %s" % (start, len(pdf.pages)))
    page_labels = Dictionary({
        "/Nums": Array([
            0, Dictionary({
                "/S": Name("/r"),  # Romano minuscolo
                "/St": 1
            }),
            (start-1), Dictionary({
                "/S": Name("/D"),  # Decimale (arabo)
                "/St": 1
            })
        ])
        })
    # Inseriamo il dizionario nel 'Catalog' (la radice del PDF)
    pdf.Root.PageLabels = page_labels

def writeOutline(pdf, outline_items):
    # scrittura indice da file
    with pdf.open_outline() as outline:
        outline.root = []
        for item in outline_items:
            # Aggiungiamo l'item all'indice del PDF
            outline.root.append(item.returnNode())

def main():
    # prendo gli argomenti
    args = getArgs().parse_args()
    file_input = args.input_pdf_file
    start = args.first_page
    file_outline = args.outline_file
    file_output = args.output_pdf_file

    # parsing indice
    outline_items = parseOutline(file_outline, start, args)

    with pikepdf.open(file_input) as pdf:
        # inserisco numerazione logica con numeri romani
        addLogicNums(pdf, start)

        # scrivo indice da file
        writeOutline(pdf, outline_items)

        # Salva il nuovo file
        pdf.save(file_output)

    print(f"File salvato con successo come: {file_output}")
    exit(0)

if __name__ == "__main__":
    main()


import unittest
from scr.main import parseOutline
from types import SimpleNamespace
from io import StringIO
from contextlib import redirect_stdout

class TestParser(unittest.TestCase):
	def setUp(self):
		self.buf = StringIO()
		self.args = SimpleNamespace(debug=False, dry=True)
		self.start = 1
	def test_noOut(self):
		# no flags, no output
		with redirect_stdout(self.buf):
			self.args.dry = False
			parseOutline("tests/sample5/index.txt", args=self.args)
			self.assertEqual("", self.buf.getvalue().strip())
	def test_wrongSyntax(self):
		# test what happens when the outline file contains a line that does not match the expected format
		with self.assertRaises(Exception) as exc:
			self.args = SimpleNamespace(debug=False, dry=False)
			parseOutline("tests/sample2/index.txt", args=self.args)
			self.assertIn("Error: line does not match the expected format:", exc.exception.args[0])
	def test_wrongPageNumber(self):
		# test what happens when the outline file contains a line with a page number that is not an integer
		with self.assertRaises(Exception) as exc:
			self.args = SimpleNamespace(debug=False, dry=False)
			parseOutline("tests/sample2/index2.txt", args=self.args)
			self.assertIn("Error: line does not match the expected format:", exc.exception.args[0])
	def test_decreasingPageNumbers(self):
		# test what happens when the outline file contains lines with page numbers that are not in increasing order
		with self.assertRaises(Exception) as exc:
			self.args = SimpleNamespace(debug=False, dry=False)
			parseOutline("tests/sample2/index3.txt", args=self.args)
			self.assertIn("Error: page numbers must be in increasing order:", exc.exception.args[0])
	def test_debug(self):
		# test the output of the debug flag
		with redirect_stdout(self.buf):
			self.args = SimpleNamespace(debug=True, dry=False)
			parseOutline("tests/sample4/index.txt", args=self.args)
			for line in self.buf.getvalue().strip().split("\n"):
				self.assertRegex(line, "title: .+, page number: \\d+, level: \\d+, prev: \\d+")
	def parseFile(self, file, ref=None):
		# parse the outline file and compare the output with the reference if provided
		if ref is None:
			ref = file
		with redirect_stdout(self.buf):
			parseOutline(file, start=self.start, args=self.args)
			with open(ref, 'r') as f:
				self.assertEqual(f.read().strip(), self.buf.getvalue().strip())
	def test_onlyChapters(self):
		# test an outline file that contains only chapters (level 0)
		self.parseFile("tests/sample3/index.txt")
	def test_singleLevel(self):
		# test an outline file that contains chapters and sections (levels 0 and 1)
		self.parseFile("tests/sample4/index.txt")
	def test_forwardJump(self):
		# test what happens when the outline file contains a line with a level that is bigger than the previous one by more than one
		self.args.dry = False
		with self.assertRaises(Exception) as exc:
			parseOutline("tests/sample6/index.txt", args=self.args)
			self.assertIn("Error: the difference between the next subsection level and this one must not be bigger than one:", exc.exception.args[0])
	def test_backwardJump(self):
		# test what happens when the outline file contains a line with a level that is smaller than the previous one by more than one
		self.args.dry = False
		parseOutline("tests/sample7/index.txt", args=self.args)
	def test_multiLevel(self):
		# test an outline file that contains any level
		self.parseFile("tests/sample5/index.txt")
	def test_multiLevelShifted(self):
		# test an outline file that contains any level, with a start different from 1
		self.start = 2
		self.parseFile("tests/sample5/index_shifted.txt", "tests/sample5/index_ref_shifted.txt")

if __name__ == '__main__':
	unittest.main()


from pikepdf import OutlineItem

class OutlineElement:
	def __init__(self, title, level, page_number, parent: OutlineElement = None):
		self.title = title
		self.level = level
		self.page_number = page_number
		self.parent = parent
		if parent is not None:
			if parent.page_number > self.page_number:
				raise Exception("Error: page numbers must be in increasing order: page title: %s, page number: %s, level: %s but parent is at page %s\n" % (title, page_number, level, parent.page_number))
		self.children = []
	
	def add_child(self, child):
		self.children.append(child)
	
	# function to create corresponding OutlineItem structure for this element and its children
	def returnNode(self):
		if self.children == []:
			return OutlineItem(self.title, self.page_number-1)
		else:
			node = OutlineItem(self.title, self.page_number-1)
			for child in self.children:
				node.children.append(child.returnNode())
			return node
	
	def __repr__(self):
		text = (self.level*" ")+str(self.page_number)+" "+self.title
		if self.children != []:
			for child in self.children:
				text += "\n" + repr(child)
		return text

# this function returns value of a Roman symbol
def romanValue(r):
	if r == 'I' or r == 'i':
		return 1
	if r == 'V' or r == 'v':
		return 5
	if r == 'X' or r == 'x':
		return 10
	if r == 'L' or r == 'l':
		return 50
	if r == 'C' or r == 'c':
		return 100
	if r == 'D' or r == 'd':
		return 500
	if r == 'M' or r == 'm':
		return 1000
	return -1

# returns decimal value of page number
def getNumber(s, start=1):
	try:
		return(int(s))+start-1	# la numerazione reale parte da start(-1 perché l'indice parte da 0)
	except ValueError:
		pass
	res = 0
	i = 0
	while i < len(s):
		
		# get value of current symbol
		s1 = romanValue(s[i])
		if s1 == -1:
			raise Exception("Error: invalid page number: %s" % s)

		# compare with the next symbol if it exists
		if i + 1 < len(s):
			s2 = romanValue(s[i + 1])

			# if current value is greater or equal, 
			# add it to result
			if s1 >= s2:
				res += s1
			else:
				# else, add the difference and 
				# skip next symbol
				res += (s2 - s1)
				i += 1
		else:
			res += s1
		i += 1

	return res


import unittest
from scr.main import writeOutline, parseOutline
import pikepdf

class TestMain(unittest.TestCase):
	def write(self, pdf_file, ref_file):
		# write the outline items to the pdf file and compare them with the reference if provided
		with pikepdf.open(pdf_file) as pdf:
			writeOutline(pdf, self.outline_items)
			with pikepdf.open(ref_file) as ref:
				self.assertEqual(pdf.Root.Outlines, ref.Root.Outlines)
	def test_toEmpty(self):
		# test wrriting an empty outline
		self.outline_items = parseOutline("tests/sample1/index.txt")
		self.write("tests/sample1/input.pdf", "tests/sample1/pdf_ref.pdf")
	def test_fromEmpty(self):
		# test writing an outline to a pdf that does not have one
		self.outline_items = parseOutline("tests/sample5/index.txt")
		self.write("tests/sample5/input.pdf", "tests/sample5/pdf_ref.pdf")
	def test_fromNonEmpty(self):
		# test writing an outline to a pdf that already has one
		self.outline_items = parseOutline("tests/sample3/index.txt")
		self.write("tests/sample3/input.pdf", "tests/sample3/pdf_ref.pdf")
	def test_NonExistingPage(self):
		# test what happens when the outline contains a page number that is bigger than the last page number of the pdf
		self.outline_items = parseOutline("tests/sample7/index.txt")
		with self.assertRaises(IndexError):
			self.write("tests/sample7/input.pdf", "tests/sample7/pdf_ref.pdf")
	
if __name__ == '__main__':
	unittest.main()